/* This code is automatically generated by genipc_server.py. Do not edit. */

#include <xmmsc/xmmsv.h>

static void
xmms_mediainfo_reader_register_ipc_commands (xmms_object_t *mediainfo_reader_object)
{
	xmms_ipc_object_register (XMMS_IPC_OBJECT_MEDIAINFO_READER, mediainfo_reader_object);


	xmms_ipc_broadcast_register (mediainfo_reader_object, XMMS_IPC_SIGNAL_MEDIAINFO_READER_STATUS);

	xmms_ipc_signal_register (mediainfo_reader_object, XMMS_IPC_SIGNAL_MEDIAINFO_READER_UNINDEXED);
}

static void
xmms_mediainfo_reader_unregister_ipc_commands (void)
{
	xmms_ipc_broadcast_unregister (XMMS_IPC_SIGNAL_MEDIAINFO_READER_STATUS);

	xmms_ipc_signal_unregister (XMMS_IPC_SIGNAL_MEDIAINFO_READER_UNINDEXED);

	xmms_ipc_object_unregister (XMMS_IPC_OBJECT_MEDIAINFO_READER);
}
