/* This code is automatically generated by genipc_server.py. Do not edit. */

#include <xmmsc/xmmsv.h>

static void
xmms_ipc_manager_register_ipc_commands (xmms_object_t *ipc_manager_object)
{
	xmms_ipc_object_register (XMMS_IPC_OBJECT_IPC_MANAGER, ipc_manager_object);


	xmms_ipc_broadcast_register (ipc_manager_object, XMMS_IPC_SIGNAL_IPC_MANAGER_CLIENT_CONNECTED);
	xmms_ipc_broadcast_register (ipc_manager_object, XMMS_IPC_SIGNAL_IPC_MANAGER_CLIENT_DISCONNECTED);

}

static void
xmms_ipc_manager_unregister_ipc_commands (void)
{
	xmms_ipc_broadcast_unregister (XMMS_IPC_SIGNAL_IPC_MANAGER_CLIENT_CONNECTED);
	xmms_ipc_broadcast_unregister (XMMS_IPC_SIGNAL_IPC_MANAGER_CLIENT_DISCONNECTED);


	xmms_ipc_object_unregister (XMMS_IPC_OBJECT_IPC_MANAGER);
}
